from deprecated.sphinx import deprecated
import pandas as pd


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_has_multiple_roots instead.",
)
def alifestd_has_multiple_roots(phylogeny_df: pd.DataFrame) -> bool:
    """Does the phylogeny two or more root organisms?

    Input dataframe is not mutated by this operation.
    """
    if "ancestor_id" in phylogeny_df.columns:
        return (phylogeny_df["ancestor_id"] == phylogeny_df["id"]).sum() >= 2
    else:
        return (
            (phylogeny_df["ancestor_list"] == "[]").sum()
            + (phylogeny_df["ancestor_list"] == "[NONE]").sum()
            + (phylogeny_df["ancestor_list"] == "[none]").sum()
            + (phylogeny_df["ancestor_list"] == "[None]").sum()
        ) >= 2



import argparse
import functools
import logging
import os
import types
import typing

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_find_root_ids import alifestd_find_root_ids
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._eval_kwargs import eval_kwargs
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned(
    insert=True,
    delete=False,
    update=True,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_add_global_root instead.",
)
def alifestd_add_global_root(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    root_attrs: typing.Mapping[str, typing.Any] = types.MappingProxyType({}),
) -> pd.DataFrame:
    """Add a new global root node that all existing roots point to.

    The new root node will have columns `id`, `ancestor_id` (if applicable),
    `ancestor_list` (if applicable), and any columns specified in
    `root_attrs`. All other columns will be NaN for the new root row.

    Parameters
    ----------
    phylogeny_df : pd.DataFrame
        Phylogeny dataframe in alife standard format.
    mutate : bool, default False
        If True, allows mutation of the input dataframe.
    root_attrs : Mapping[str, Any], default {}
        Column values to set on the new global root row, e.g.,
        ``{"origin_time": 0.0, "taxon_label": "root"}``.

        Keys ``"id"``, ``"ancestor_id"``, and ``"ancestor_list"`` are
        reserved and may not be specified; a `ValueError` is raised if
        any are present.

    Returns
    -------
    pd.DataFrame
        The phylogeny dataframe with a new global root added.

    Raises
    ------
    ValueError
        If `root_attrs` contains reserved keys.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    bad_keys = {"id", "ancestor_id", "ancestor_list"} & root_attrs.keys()
    if bad_keys:
        raise ValueError(
            f"root_attrs must not contain reserved keys {bad_keys}; "
            "these are set automatically"
        )

    # Create new root id
    new_root_id = phylogeny_df["id"].max() + 1 if not phylogeny_df.empty else 0

    # Build the new root row
    new_root = {
        "id": new_root_id,
        **(
            {"ancestor_id": new_root_id}
            if "ancestor_id" in phylogeny_df
            else {}
        ),
        **(
            {"ancestor_list": "[none]"}
            if "ancestor_list" in phylogeny_df
            else {}
        ),
        **root_attrs,
    }

    # Point existing roots to the new global root
    root_ids = alifestd_find_root_ids(phylogeny_df)
    root_mask = phylogeny_df["id"].isin(root_ids)

    if "ancestor_id" in phylogeny_df:
        phylogeny_df.loc[root_mask, "ancestor_id"] = new_root_id

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[root_mask, "ancestor_list"] = f"[{new_root_id}]"

    # Append the new root row (vertical concat; mismatched columns get NaN)
    new_root_df = pd.DataFrame([new_root])
    res = pd.concat([phylogeny_df, new_root_df], axis=0, ignore_index=True)

    return res


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Add a global root node that all existing roots point to.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_add_global_root",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--root-attr",
        action="append",
        dest="root_attrs",
        type=str,
        default=[],
        help=(
            "Column value to set on the new root row. "
            "Provide as 'key=value'. "
            "Specify multiple attrs by using this flag multiple times. "
            "Arguments will be evaluated as Python expressions. "
            "Example: --root-attr 'origin_time=0.0'"
        ),
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_add_global_root",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_add_global_root,
                    root_attrs=eval_kwargs(args.root_attrs),
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
        )



import os
import types

from downstream import dstream, dsurf
import pandas as pd
import pytest

from hstrat import hstrat

assets_path = os.path.join(
    os.path.dirname(__file__), os.pardir, "test_dataframe", "assets"
)


def _make_dstream_df(
    surfaces,
    algo,
    stratum_differentia_bit_width,
    dstream_T_bitwidth=32,
):
    """Build a pandas DataFrame with dstream metadata from surfaces."""
    rows = []
    for surf in surfaces:
        hex_string = hstrat.surf_to_hex(
            surf, dstream_T_bitwidth=dstream_T_bitwidth
        )
        rows.append(
            {
                "data_hex": hex_string,
                "dstream_algo": f"dstream.{algo.__name__.split('.')[-1]}",
                "dstream_storage_bitoffset": dstream_T_bitwidth,
                "dstream_storage_bitwidth": surf.S
                * stratum_differentia_bit_width,
                "dstream_T_bitoffset": 0,
                "dstream_T_bitwidth": dstream_T_bitwidth,
                "dstream_S": surf.S,
            },
        )
    return pd.DataFrame(rows)


# -------------------------------------------------------------------
# Round-trip tests: create surfaces, serialize to df, deserialize
# -------------------------------------------------------------------


@pytest.mark.parametrize(
    "algo",
    [
        dstream.steady_algo,
        dstream.stretched_algo,
        dstream.tilted_algo,
    ],
)
@pytest.mark.parametrize("S", [8, 16])
@pytest.mark.parametrize("n", [0, 1, 5, 25, 100])
@pytest.mark.parametrize("stratum_differentia_bit_width", [1, 8, 64])
def test_round_trip(
    algo: types.ModuleType,
    S: int,
    n: int,
    stratum_differentia_bit_width: int,
):
    """Surfaces serialized to a dstream df and deserialized back into
    an assemblage should preserve retained ranks and differentia."""
    surfaces = [
        hstrat.HereditaryStratigraphicSurface(
            dsurf.Surface(algo, S),
            predeposit_strata=0,
            stratum_differentia_bit_width=stratum_differentia_bit_width,
        )
        for _ in range(3)
    ]
    for surf in surfaces:
        surf.DepositStrata(n)

    df = _make_dstream_df(surfaces, algo, stratum_differentia_bit_width)
    assemblage = hstrat.assemblage_from_dstream_df(df)

    for surf, specimen in zip(surfaces, assemblage.BuildSpecimens()):
        retained_ranks = list(surf.IterRetainedRanks())
        retained_diff = list(surf.IterRetainedDifferentia())

        specimen_ranks = list(specimen.IterRetainedRanks())
        specimen_diff = list(specimen.IterRetainedDifferentia())

        assert specimen_ranks == retained_ranks
        assert specimen_diff == retained_diff


def test_single_surface():
    """A single surface should produce a one-specimen assemblage."""
    algo = dstream.steady_algo
    S = 8
    bit_width = 8
    surf = hstrat.HereditaryStratigraphicSurface(
        dsurf.Surface(algo, S),
        predeposit_strata=0,
        stratum_differentia_bit_width=bit_width,
    )
    surf.DepositStrata(10)

    df = _make_dstream_df([surf], algo, bit_width)
    assemblage = hstrat.assemblage_from_dstream_df(df)

    specimens = list(assemblage.BuildSpecimens())
    assert len(specimens) == 1
    assert list(specimens[0].IterRetainedRanks()) == list(
        surf.IterRetainedRanks()
    )
    assert list(specimens[0].IterRetainedDifferentia()) == list(
        surf.IterRetainedDifferentia()
    )


def test_multiple_surfaces():
    """Multiple surfaces should produce corresponding specimens."""
    algo = dstream.steady_algo
    S = 8
    bit_width = 8
    surfaces = []
    for i in range(5):
        surf = hstrat.HereditaryStratigraphicSurface(
            dsurf.Surface(algo, S),
            predeposit_strata=0,
            stratum_differentia_bit_width=bit_width,
        )
        surf.DepositStrata(10 + i * 5)
        surfaces.append(surf)

    df = _make_dstream_df(surfaces, algo, bit_width)
    assemblage = hstrat.assemblage_from_dstream_df(df)

    specimens = list(assemblage.BuildSpecimens())
    assert len(specimens) == len(surfaces)


# -------------------------------------------------------------------
# Error handling tests
# -------------------------------------------------------------------


def test_missing_column_raises():
    """Raises ValueError when a required column is absent."""
    algo = dstream.steady_algo
    S = 8
    bit_width = 8
    surf = hstrat.HereditaryStratigraphicSurface(
        dsurf.Surface(algo, S),
        predeposit_strata=0,
        stratum_differentia_bit_width=bit_width,
    )
    surf.DepositStrata(10)

    df = _make_dstream_df([surf], algo, bit_width)
    stripped = df.drop(columns=["data_hex"])
    with pytest.raises(ValueError, match="missing required columns"):
        hstrat.assemblage_from_dstream_df(stripped)


def test_missing_multiple_columns_raises():
    """Raises ValueError listing all missing columns."""
    algo = dstream.steady_algo
    S = 8
    bit_width = 8
    surf = hstrat.HereditaryStratigraphicSurface(
        dsurf.Surface(algo, S),
        predeposit_strata=0,
        stratum_differentia_bit_width=bit_width,
    )
    surf.DepositStrata(10)

    df = _make_dstream_df([surf], algo, bit_width)
    stripped = df.drop(columns=["data_hex", "dstream_algo"])
    with pytest.raises(ValueError, match="missing required columns"):
        hstrat.assemblage_from_dstream_df(stripped)


# -------------------------------------------------------------------
# Smoke test with CSV test asset
# -------------------------------------------------------------------


def test_from_packed_csv():
    """Smoke test: deserialize assemblage from packed_simple.csv asset."""
    df = pd.read_csv(f"{assets_path}/packed_simple.csv")
    assemblage = hstrat.assemblage_from_dstream_df(df)
    specimens = list(assemblage.BuildSpecimens())
    assert len(specimens) == 2  # packed_simple.csv has 2 rows


# -------------------------------------------------------------------
# progress_wrap parameter test
# -------------------------------------------------------------------


def test_progress_wrap():
    """The progress_wrap callable is applied to the row iterator."""
    algo = dstream.steady_algo
    S = 8
    bit_width = 8

    surfs = [
        hstrat.HereditaryStratigraphicSurface(
            dsurf.Surface(algo, S),
            predeposit_strata=0,
            stratum_differentia_bit_width=bit_width,
        )
        for _ in range(3)
    ]
    for s in surfs:
        s.DepositStrata(10)

    df = _make_dstream_df(surfs, algo, bit_width)

    call_count = 0

    def counting_wrap(it):
        nonlocal call_count
        call_count += 1
        return it

    hstrat.assemblage_from_dstream_df(df, progress_wrap=counting_wrap)
    assert call_count == 1


# -------------------------------------------------------------------
# Differentia bit width preserved test
# -------------------------------------------------------------------


@pytest.mark.parametrize("stratum_differentia_bit_width", [1, 2, 8, 64])
def test_differentia_bit_width_preserved(
    stratum_differentia_bit_width,
):
    """Assemblage specimens have correct differentia bit width."""
    algo = dstream.steady_algo
    S = 8
    surf = hstrat.HereditaryStratigraphicSurface(
        dsurf.Surface(algo, S),
        predeposit_strata=0,
        stratum_differentia_bit_width=stratum_differentia_bit_width,
    )
    surf.DepositStrata(10)

    df = _make_dstream_df([surf], algo, stratum_differentia_bit_width)
    assemblage = hstrat.assemblage_from_dstream_df(df)
    specimen = next(assemblage.BuildSpecimens())
    assert (
        specimen.GetStratumDifferentiaBitWidth()
        == stratum_differentia_bit_width
    )



import math
import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_find_root_ids,
    alifestd_has_multiple_roots,
    alifestd_make_balanced_bifurcating,
    alifestd_make_comb,
    alifestd_make_empty,
    alifestd_mark_colless_like_index_mdm_asexual,
    alifestd_mark_num_children_asexual,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)
from hstrat._auxiliary_lib._alifestd_mark_colless_like_index_mdm_asexual import (
    _colless_like_fast_path,
    _colless_like_slow_path,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_colless_like_index_mdm_asexual(phylogeny_df)

    assert alifestd_validate(result)
    assert original.equals(phylogeny_df)

    # Colless-like index should be non-negative
    assert all(result["colless_like_index_mdm"] >= 0)

    # Leaf nodes should have Colless-like index 0
    leaf_ids = [*alifestd_find_leaf_ids(phylogeny_df)]
    for leaf_id in leaf_ids:
        val = result[result["id"] == leaf_id][
            "colless_like_index_mdm"
        ].squeeze()
        assert val == 0.0

    # Root should have the highest or equal index
    if not alifestd_has_multiple_roots(phylogeny_df):
        (root_id,) = alifestd_find_root_ids(phylogeny_df)
        root_val = result[result["id"] == root_id][
            "colless_like_index_mdm"
        ].squeeze()
        assert root_val >= 0
        assert root_val == result["colless_like_index_mdm"].max()


def test_empty():
    res = alifestd_mark_colless_like_index_mdm_asexual(
        alifestd_make_empty(),
    )
    assert "colless_like_index_mdm" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_chain(mutate: bool):
    """Test a simple chain: 0 -> 1 -> 2.

    All nodes are unifurcating (<=1 child), so balance = 0.
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[1, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[2, "colless_like_index_mdm"] == 0.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_balanced(mutate: bool):
    r"""Test a balanced bifurcating tree.

          0
         / \
        1   2

    Both children are leaves with identical f-size = f(0) = ln(e) = 1.
    MDM(1, 1) = 0.  So index at root = 0.
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "colless_like_index_mdm"] == pytest.approx(
        0.0,
    )
    assert result_df.loc[1, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[2, "colless_like_index_mdm"] == 0.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_imbalanced(mutate: bool):
    r"""Test an imbalanced bifurcating tree.

          0
         / \
        1   2
           / \
          3   4

    f-sizes:
    - Leaf (deg 0): f(0) = ln(e) = 1.0
    - Node 2 (deg 2): f(2) + f_size(3) + f_size(4)
                     = ln(2+e) + 1.0 + 1.0 = ln(2+e) + 2.0
    - Node 1 (leaf): f_size = 1.0

    Node 2 balance: children 3, 4 have equal f-sizes (1.0, 1.0)
                    MDM(1,1) = 0
    Node 0 balance: children have f-sizes 1.0 and ln(2+e)+2.0
                    diff = ln(2+e) + 1.0
                    MDM = (1/2) * (diff/2 + diff/2) = diff/2
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[2]", "[2]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have index 0
    assert result_df.loc[1, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[3, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[4, "colless_like_index_mdm"] == 0.0

    # Node 2: balanced children -> 0
    assert result_df.loc[2, "colless_like_index_mdm"] == pytest.approx(
        0.0,
    )

    # Node 0: MDM of (1.0, ln(2+e)+2.0)
    fsize_1 = math.log(0 + math.e)  # 1.0
    fsize_2 = math.log(2 + math.e) + 2 * math.log(0 + math.e)  # ln(2+e) + 2.0
    diff = abs(fsize_2 - fsize_1)
    expected_bal = diff / 2.0  # MDM of two values
    assert result_df.loc[0, "colless_like_index_mdm"] == pytest.approx(
        expected_bal,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_balanced(mutate: bool):
    r"""Test a tree with balanced polytomy.

          0
        / | \
       1  2  3

    All children are leaves with f-size = 1.0.
    MDM(1, 1, 1) = 0. Index at root = 0.
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[1, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[2, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[3, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[0, "colless_like_index_mdm"] == pytest.approx(
        0.0,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_imbalanced(mutate: bool):
    r"""Test a tree with imbalanced polytomy.

            0
          / | \
         1  2  3
              / \
             4   5

    f-sizes:
    - Leaves 1, 2, 4, 5: f(0) = 1.0
    - Node 3: f(2) + 1.0 + 1.0 = ln(2+e) + 2.0
    - Children of 0: 1.0, 1.0, ln(2+e)+2.0

    Sorted children f-sizes: [1.0, 1.0, ln(2+e)+2.0]
    Median = 1.0 (middle value of 3)
    MDM = (1/3) * (|1-1| + |1-1| + |ln(2+e)+2-1|)
        = (1/3) * (ln(2+e) + 1.0)
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[3]",
                "[3]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves
    for leaf in [1, 2, 4, 5]:
        assert result_df.loc[leaf, "colless_like_index_mdm"] == 0.0

    # Node 3: balanced children -> 0
    assert result_df.loc[3, "colless_like_index_mdm"] == pytest.approx(
        0.0,
    )

    # Node 0: MDM of [1.0, 1.0, ln(2+e)+2.0], median=1.0
    fsize_3 = math.log(2 + math.e) + 2 * math.log(0 + math.e)
    expected_bal = (abs(1.0 - 1.0) + abs(1.0 - 1.0) + abs(fsize_3 - 1.0)) / 3
    assert result_df.loc[0, "colless_like_index_mdm"] == pytest.approx(
        expected_bal,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_non_contiguous_ids(mutate: bool):
    """Test with non-contiguous IDs to exercise slow path."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50],
            "ancestor_list": [
                "[None]",
                "[10]",
                "[10]",
                "[30]",
                "[30]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Same structure as test_simple_bifurcating_imbalanced
    assert result_df.loc[20, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[40, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[50, "colless_like_index_mdm"] == 0.0
    assert result_df.loc[30, "colless_like_index_mdm"] == pytest.approx(
        0.0,
    )

    fsize_20 = math.log(0 + math.e)
    fsize_30 = math.log(2 + math.e) + 2 * math.log(0 + math.e)
    diff = abs(fsize_30 - fsize_20)
    expected_bal = diff / 2.0
    assert result_df.loc[10, "colless_like_index_mdm"] == pytest.approx(
        expected_bal,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_fast_slow_path_agreement(mutate: bool):
    """Fast and slow paths should produce the same values."""
    phylogeny_df_nc = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50, 60, 70],
            "ancestor_list": [
                "[None]",
                "[10]",
                "[10]",
                "[10]",
                "[10]",
                "[50]",
                "[50]",
            ],
        }
    )
    phylogeny_df_c = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[0]",
                "[4]",
                "[4]",
            ],
        }
    )
    result_nc = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df_nc,
        mutate=mutate,
    )
    result_c = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df_c,
        mutate=mutate,
    )

    result_nc.index = result_nc["id"]
    result_c.index = result_c["id"]
    assert result_nc.loc[10, "colless_like_index_mdm"] == pytest.approx(
        result_c.loc[0, "colless_like_index_mdm"],
    )


def test_symmetric_tree_is_zero():
    """Fully symmetric trees should have Colless-like index 0."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
        }
    )
    result_df = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df,
    )
    assert result_df["colless_like_index_mdm"].to_list() == pytest.approx(
        [0.0] * len(result_df),
    )

    # Symmetric trifurcating tree (3 leaves)
    phylogeny_df2 = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]"],
        }
    )
    result_df2 = alifestd_mark_colless_like_index_mdm_asexual(
        phylogeny_df2,
    )
    assert result_df2["colless_like_index_mdm"].to_list() == pytest.approx(
        [0.0] * len(result_df2),
    )


@pytest.mark.parametrize(
    "phylogeny_df, expected_cl",
    [
        (alifestd_make_balanced_bifurcating(2), 0.0),
        (alifestd_make_balanced_bifurcating(3), 0.0),
        (alifestd_make_balanced_bifurcating(4), 0.0),
        (alifestd_make_comb(3), 2.551444713932051 / 2),
        (alifestd_make_comb(4), 7.654334141796154 / 2),
        (alifestd_make_comb(5), 15.308668283592308 / 2),
    ],
)
def test_against_r_treebalance_colless_like(
    phylogeny_df: pd.DataFrame, expected_cl: float
):
    """Test Colless-like index against values derived from R
    treebalance::collesslikeI(, f.size="ln", dissim="mdm").

    Note: R treebalance 1.0.x has an operator precedence bug in its MDM
    dissimilarity where ``return(sum(abs(x - median(x))))/length(x)``
    puts the division outside the return, computing sum instead of mean
    absolute deviation from median. For bifurcating trees (all nodes
    k=2), the R values are exactly 2x the correct MDM values, so we
    divide by 2 to get the correct expected values per Mir et al. (2018).
    """
    result_df = alifestd_mark_colless_like_index_mdm_asexual(phylogeny_df)
    result_df.index = result_df["id"]
    root_id = result_df[result_df["id"] == result_df["ancestor_id"]][
        "id"
    ].iloc[0]
    assert result_df.loc[root_id, "colless_like_index_mdm"] == pytest.approx(
        expected_cl,
    )


def test_colless_like_trifurcation_mixed():
    """Test Colless-like index for a tree with trifurcation.

    Correct MDM at root: children f-sizes = [1.0, 1.0, ln(2+e)+2.0]
    Median = 1.0 (middle of 3 sorted values)
    MDM = (1/3) * (0 + 0 + |ln(2+e)+2.0 - 1.0|) = (ln(2+e)+1.0)/3
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[3]",
                "[3]",
            ],
        }
    )
    result_df = alifestd_mark_colless_like_index_mdm_asexual(phylogeny_df)
    result_df.index = result_df["id"]

    r_buggy_value = 2.551444713932051
    expected_root = r_buggy_value / 3.0
    assert result_df.loc[0, "colless_like_index_mdm"] == pytest.approx(
        expected_root,
    )


@pytest.mark.parametrize(
    "phylogeny_df, expected_cl",
    [
        (
            pd.DataFrame(
                {
                    "id": range(9),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[6]",
                        "[6]",
                    ],
                }
            ),
            2.5514447139320513,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(9),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[4]",
                        "[4]",
                    ],
                }
            ),
            2.5514447139320513,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(11),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[5]",
                        "[5]",
                    ],
                }
            ),
            2.5514447139320513,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(11),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[4]",
                        "[4]",
                    ],
                }
            ),
            2.5514447139320513,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(13),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[4]",
                        "[4]",
                        "[6]",
                        "[6]",
                        "[8]",
                        "[8]",
                    ],
                }
            ),
            6.378611784830128,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(17),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[4]",
                        "[4]",
                        "[5]",
                        "[5]",
                        "[6]",
                        "[6]",
                        "[10]",
                        "[10]",
                    ],
                }
            ),
            3.827167070898077,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(15),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[5]",
                        "[5]",
                        "[6]",
                        "[6]",
                        "[9]",
                        "[9]",
                        "[10]",
                        "[10]",
                    ],
                }
            ),
            7.654334141796154,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(8),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[0]",
                        "[3]",
                        "[3]",
                        "[5]",
                        "[5]",
                    ],
                }
            ),
            2.976685499587393,
        ),
        (
            pd.DataFrame(
                {
                    "id": range(10),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[7]",
                        "[7]",
                    ],
                }
            ),
            1.955332880583737,
        ),
    ],
)
def test_against_r_nontrivial_trees_colless_like(
    phylogeny_df: pd.DataFrame, expected_cl: float
):
    """Test Colless-like index against values derived from R
    treebalance::collesslikeI for nontrivial trees, corrected for the
    R MDM bug."""
    result_df = alifestd_mark_colless_like_index_mdm_asexual(phylogeny_df)
    result_df.index = result_df["id"]
    root_id = result_df[result_df["id"] == result_df["ancestor_id"]][
        "id"
    ].iloc[0]
    assert result_df.loc[root_id, "colless_like_index_mdm"] == pytest.approx(
        expected_cl,
    )


def test_fast_slow_path_direct_comparison():
    """Directly call fast and slow paths and compare results."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[0]",
                "[4]",
                "[4]",
            ],
        }
    )
    phylogeny_df = alifestd_try_add_ancestor_id_col(
        phylogeny_df,
        mutate=True,
    )
    phylogeny_df = alifestd_mark_num_children_asexual(
        phylogeny_df,
        mutate=True,
    )
    fast_result = _colless_like_fast_path(
        phylogeny_df["ancestor_id"].to_numpy(),
        0,
    )
    slow_result = _colless_like_slow_path(phylogeny_df.copy(), "mdm")

    np.testing.assert_allclose(fast_result, slow_result)



import logging

from hstrat._auxiliary_lib import begin_prod_logging, get_hstrat_version


def test_smoke():
    begin_prod_logging()


def test_logs_version(caplog):
    with caplog.at_level(logging.INFO):
        begin_prod_logging()

    assert any(
        get_hstrat_version() in record.message for record in caplog.records
    )



import typing

from deprecated.sphinx import deprecated
import numpy as np
import opytional as opyt
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit


@jit(nopython=True)
def _alifestd_find_pair_mrca_id_asexual_fast_path(
    ancestor_ids: np.ndarray,
    first: int,
    second: int,
) -> int:
    """Find the MRCA of two taxa in a contiguous, topologically sorted
    asexual phylogeny.

    Parameters
    ----------
    ancestor_ids : np.ndarray
        Array of ancestor ids, where index corresponds to taxon id.

        Must be topologically sorted with contiguous ids.
    first : int
        First taxon id.
    second : int
        Second taxon id.

    Returns
    -------
    int
        The id of the most recent common ancestor, or -1 if no common
        ancestor exists.
    """
    # walk both lineages towards the root
    # by advancing the deeper (higher-id) node first
    a, b = first, second
    while a != b:
        a, b = min(a, b), max(a, b)
        next_b = ancestor_ids[b]
        if next_b == b:
            assert a != b
            return -1  # b is a root, disjoint trees
        b = next_b
    assert a == b
    return a


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_find_pair_mrca_id_asexual instead.",
)
def alifestd_find_pair_mrca_id_asexual(
    phylogeny_df: pd.DataFrame,
    first: int,
    second: int,
    *,
    mutate: bool = False,
    is_topologically_sorted: typing.Optional[bool] = None,
    has_contiguous_ids: typing.Optional[bool] = None,
) -> typing.Optional[int]:
    """Find the Most Recent Common Ancestor of two taxa.

    Parameters
    ----------
    phylogeny_df : pd.DataFrame
        Phylogeny in alife standard format.
    first : int
        First taxon id.
    second : int
        Second taxon id.
    mutate : bool, default False
        If True, allows in-place modification of `phylogeny_df`.
    is_topologically_sorted : bool, optional
        If provided, skips the topological sort check. If None
        (default), the check is performed automatically.
    has_contiguous_ids : bool, optional
        If provided, skips the contiguous ids check. If None (default),
        the check is performed automatically.

    Returns
    -------
    int or None
        The id of the most recent common ancestor, or None if no common
        ancestor exists.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    if "ancestor_id" not in phylogeny_df.columns:
        raise ValueError(
            "alifestd_find_pair_mrca_id_asexual requires ancestor_id column",
        )

    if not opyt.or_else(
        is_topologically_sorted,
        lambda: alifestd_is_topologically_sorted(phylogeny_df),
    ):
        raise NotImplementedError(
            "topologically unsorted rows not yet supported",
        )

    if not opyt.or_else(
        has_contiguous_ids,
        lambda: alifestd_has_contiguous_ids(phylogeny_df),
    ):
        raise NotImplementedError("non-contiguous ids not yet supported")

    ancestor_ids = phylogeny_df["ancestor_id"].to_numpy()
    result = _alifestd_find_pair_mrca_id_asexual_fast_path(
        ancestor_ids,
        first,
        second,
    )
    return None if result == -1 else int(result)



import argparse
import functools
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._add_bool_arg import add_bool_arg
from ._alifestd_assign_contiguous_ids_polars import (
    alifestd_assign_contiguous_ids_polars,
)
from ._alifestd_collapse_unifurcations import _collapse_unifurcations
from ._alifestd_has_contiguous_ids_polars import (
    alifestd_has_contiguous_ids_polars,
)
from ._alifestd_is_topologically_sorted_polars import (
    alifestd_is_topologically_sorted_polars,
)
from ._alifestd_topological_sensitivity_warned_polars import (
    alifestd_topological_sensitivity_warned_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned_polars(
    insert=False,
    delete=True,
    update=True,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_collapse_unifurcations_polars instead.",
)
def alifestd_collapse_unifurcations_polars(
    phylogeny_df: pl.DataFrame,
) -> pl.DataFrame:
    """Pare record to bypass organisms with one ancestor and one descendant.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_collapse_unifurcations :
        Pandas-based implementation.
    """
    schema_names = phylogeny_df.lazy().collect_schema().names()

    if "ancestor_list" in schema_names:
        raise NotImplementedError

    logging.info(
        "- alifestd_collapse_unifurcations_polars: "
        "collecting original ids...",
    )
    original_ids = (
        phylogeny_df.lazy().select("id").collect().to_series().to_numpy()
    )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: "
        "checking contiguous ids...",
    )
    if not alifestd_has_contiguous_ids_polars(phylogeny_df):
        logging.info(
            "- alifestd_collapse_unifurcations_polars: assigning ids...",
        )
        phylogeny_df = alifestd_assign_contiguous_ids_polars(phylogeny_df)

    logging.info(
        "- alifestd_collapse_unifurcations_polars: "
        "checking topological sort...",
    )
    if not alifestd_is_topologically_sorted_polars(phylogeny_df):
        raise NotImplementedError(
            "polars topological sort not yet implemented",
        )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: "
        "collecting ancestor_ids...",
    )
    ancestor_ids = (
        phylogeny_df.lazy()
        .select("ancestor_id")
        .collect()
        .to_series()
        .to_numpy()
    )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: calculating reindex...",
    )
    keep_filter, ancestor_ids = _collapse_unifurcations(
        ancestor_ids.copy(),  # must copy to remove read-only flag...
        # ... for numba compatibility
    )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: applying reindex...",
    )
    return phylogeny_df.filter(keep_filter).with_columns(
        id=original_ids[keep_filter],
        ancestor_id=original_ids[ancestor_ids[keep_filter]],
    )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Pare record to bypass organisms with one ancestor and one descendant.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_collapse_unifurcations :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_collapse_unifurcations_polars",
        dfcli_version=get_hstrat_version(),
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_collapse_unifurcations_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=functools.partial(
                    alifestd_collapse_unifurcations_polars,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



import os
import pathlib
import subprocess

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_surface_test_drive_cli_help():
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            "--help",
        ],
        check=True,
    )


def test_surface_test_drive_cli_version():
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            "--version",
        ],
        check=True,
    )


def test_surface_test_drive_cli_csv():
    output_file = "/tmp/hstrat_surface_test_drive.csv"  # nosec B108
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            output_file,
        ],
        check=True,
        input=f"{assets}/nk_ecoeaselection.csv".encode(),
    )
    assert os.path.exists(output_file)


def test_surface_test_drive_cli_parquet():
    output_file = "/tmp/hstrat_surface_test_drive.pqt"  # nosec B108
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            output_file,
            "--shrink-dtypes",
        ],
        check=True,
        input=f"{assets}/nk_ecoeaselection.csv".encode(),
    )
    assert os.path.exists(output_file)


def test_surface_test_drive_cli_flags():
    output_file = "/tmp/hstrat_surface_test_drive_flags.pqt"  # nosec B108
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            output_file,
            "--shrink-dtypes",
            "--stratum-differentia-bit-width",
            "8",
            "--dstream-algo",
            "dstream.primed_4pad0_tiltedxtc_algo",
            "--dstream-S",
            "36",
            "--dstream-T-bitwidth",
            "64",
        ],
        check=True,
        input=f"{assets}/nk_ecoeaselection.csv".encode(),
    )
    assert os.path.exists(output_file)


def test_surface_test_drive_cli_pipe_reconstruct():
    output_file = (
        "/tmp/hstrat_surface_test_drive_reconstruct.pqt"  # nosec B108
    )
    pathlib.Path(output_file).unlink(missing_ok=True)
    td = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_test_drive",
            "/dev/stdout",
            "--shrink-dtypes",
            "--output-filetype",
            "parquet",
            "--write-kwarg",
            "compression='uncompressed'",
        ],
        capture_output=True,
        check=True,
        input=f"{assets}/nk_ecoeaselection.csv".encode(),
    )
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_build_tree",
            output_file,
            "--stdin",
            "--input-filetype",
            "parquet",
        ],
        check=True,
        input=td.stdout,
    )
    assert os.path.exists(output_file)



import os
import subprocess

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_surface_validate_trie_cli_help():
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            "--help",
        ],
        check=True,
    )


def test_surface_validate_trie_cli_version():
    subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            "--version",
        ],
        check=True,
    )


def test_surface_validate_trie_cli_csv():
    """Validates a well-formed trie CSV and prints violation count."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie.csv",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    assert result.stdout.strip() == "0"


def test_surface_validate_trie_cli_parquet():
    """Validates a well-formed trie from a Parquet file."""
    trie_pqt = "/tmp/hstrat_validate_trie_input.pqt"  # nosec B108
    import polars as pl

    pl.read_csv(f"{assets}/trie.csv").write_parquet(trie_pqt)

    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            trie_pqt,
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    assert result.stdout.strip() == "0"


def test_surface_validate_trie_cli_long():
    """Validates a late-divergence trie (lrc=47, mrca=55)."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie_long.csv",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    assert result.stdout.strip() == "0"


def test_surface_validate_trie_cli_fails_without_metadata():
    """CLI exits non-zero when dstream metadata columns are absent."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie_no_meta.csv",
        ],
    )
    assert result.returncode != 0


def test_surface_validate_trie_cli_fails_on_invalid_trie():
    """CLI exits non-zero when trie has MRCA rank violations."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie_long_invalid.csv",
            "--max-violations",
            "0",
        ],
    )
    assert result.returncode != 0


def test_surface_validate_trie_cli_max_num_checks_zero():
    """With --max-num-checks 0, no leaf-pair checks run; always passes."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie_long_invalid.csv",
            "--max-num-checks",
            "0",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    assert result.stdout.strip() == "0"


def test_surface_validate_trie_cli_max_violations():
    """--max-violations controls the threshold before exit-on-error."""
    result = subprocess.run(  # nosec B603
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_validate_trie",
            f"{assets}/trie_long_invalid.csv",
            "--max-violations",
            "9999",
        ],
        check=True,
        capture_output=True,
        text=True,
    )
    # violation count is printed even when within tolerance
    assert result.stdout.strip().isdigit()


def test_surface_validate_trie_cli_seed():
    """--seed flag is accepted and produces reproducible results."""
    results = []
    for _ in range(2):
        r = subprocess.run(  # nosec B603
            [
                "python3",
                "-m",
                "hstrat.dataframe.surface_validate_trie",
                f"{assets}/trie.csv",
                "--seed",
                "42",
            ],
            check=True,
            capture_output=True,
            text=True,
        )
        results.append(r.stdout.strip())
    assert results[0] == results[1]



import os
import typing

import pandas as pd
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_try_add_ancestor_id_col,
    alifestd_try_add_ancestor_list_col,
)
from hstrat._auxiliary_lib._alifestd_make_ancestor_list_col_polars import (
    alifestd_make_ancestor_list_col_polars,
)
from hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col_polars import (
    alifestd_try_add_ancestor_list_col_polars,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_adds_col(
    phylogeny_df, apply: typing.Callable
):
    """Verify ancestor_list is added when ancestor_id is present."""
    phylogeny_df_pd = alifestd_try_add_ancestor_id_col(phylogeny_df.copy())
    df_prepared = pl.from_pandas(phylogeny_df_pd[["id", "ancestor_id"]])
    df = apply(df_prepared)

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert "ancestor_list" in result.columns
    assert len(result) == len(df_prepared)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_correctness(
    phylogeny_df, apply: typing.Callable
):
    """Verify generated ancestor_list matches expected values."""
    phylogeny_df_pd = alifestd_try_add_ancestor_id_col(phylogeny_df.copy())
    df_prepared = pl.from_pandas(phylogeny_df_pd[["id", "ancestor_id"]])
    df = apply(df_prepared)

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    expected = alifestd_make_ancestor_list_col_polars(
        result["id"],
        result["ancestor_id"],
        root_ancestor_token="none",
    )
    assert result["ancestor_list"].to_list() == expected.to_list()


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_matches_pandas(
    phylogeny_df, apply: typing.Callable
):
    """Verify polars result matches pandas result."""
    phylogeny_df_pd = alifestd_try_add_ancestor_id_col(phylogeny_df.copy())

    # Run pandas version (drop ancestor_list first to force re-creation)
    pd_input = phylogeny_df_pd.drop(columns="ancestor_list")
    result_pd = alifestd_try_add_ancestor_list_col(pd_input)

    # Run polars version
    pl_input = apply(pl.from_pandas(pd_input))
    result_pl = (
        alifestd_try_add_ancestor_list_col_polars(pl_input).lazy().collect()
    )

    assert result_pd["ancestor_list"].tolist() == (
        result_pl["ancestor_list"].to_list()
    )


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_already_has_col(
    apply: typing.Callable,
):
    """Verify no change when ancestor_list column already exists."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
                "ancestor_list": ["[none]", "[0]", "[1]"],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert result["ancestor_list"].to_list() == ["[none]", "[0]", "[1]"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_no_ancestor_id(
    apply: typing.Callable,
):
    """Verify no change when ancestor_id column is missing."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_list": ["[none]", "[0]", "[1]"],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert result["ancestor_list"].to_list() == ["[none]", "[0]", "[1]"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_neither_col(
    apply: typing.Callable,
):
    """Verify no change when both ancestor_id and ancestor_list are missing."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "origin_time": [0.0, 1.0, 2.0],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert "ancestor_list" not in result.columns


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_simple_chain(
    apply: typing.Callable,
):
    """Test a simple chain: 0 -> 1 -> 2."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert "ancestor_list" in result.columns
    assert result["ancestor_list"].to_list() == ["[none]", "[0]", "[1]"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_simple_tree(
    apply: typing.Callable,
):
    """Test a simple tree.

    Tree structure:
        0 (root)
        +-- 1
        |   +-- 3
        |   +-- 4
        +-- 2
    """
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3, 4],
                "ancestor_id": [0, 0, 0, 1, 1],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert result["ancestor_list"].to_list() == [
        "[none]",
        "[0]",
        "[0]",
        "[1]",
        "[1]",
    ]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_root_ancestor_token(
    apply: typing.Callable,
):
    """Test custom root_ancestor_token."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        )
    )

    result = (
        alifestd_try_add_ancestor_list_col_polars(
            df, root_ancestor_token="None"
        )
        .lazy()
        .collect()
    )
    assert result["ancestor_list"].to_list() == ["[None]", "[0]", "[1]"]

    result2 = (
        alifestd_try_add_ancestor_list_col_polars(df, root_ancestor_token="")
        .lazy()
        .collect()
    )
    assert result2["ancestor_list"].to_list() == ["[]", "[0]", "[1]"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_single_node(
    apply: typing.Callable,
):
    """A single root node."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0],
                "ancestor_id": [0],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert result["ancestor_list"].to_list() == ["[none]"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_empty(
    apply: typing.Callable,
):
    """Verify empty dataframe."""
    df = apply(
        pl.DataFrame(
            {"id": [], "ancestor_id": []},
            schema={"id": pl.Int64, "ancestor_id": pl.Int64},
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert "ancestor_list" in result.columns
    assert result.is_empty()


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_preserves_columns(
    apply: typing.Callable,
):
    """Verify original columns are preserved."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
                "origin_time": [0.0, 1.0, 2.0],
                "taxon_label": ["a", "b", "c"],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert "ancestor_list" in result.columns
    assert "origin_time" in result.columns
    assert "taxon_label" in result.columns
    assert result["origin_time"].to_list() == [0.0, 1.0, 2.0]
    assert result["taxon_label"].to_list() == ["a", "b", "c"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_idempotent(
    apply: typing.Callable,
):
    """Calling twice should produce the same result."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        )
    )

    result1 = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()
    result2 = (
        alifestd_try_add_ancestor_list_col_polars(apply(result1))
        .lazy()
        .collect()
    )

    assert (
        result1["ancestor_list"].to_list()
        == result2["ancestor_list"].to_list()
    )


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_try_add_ancestor_list_col_polars_two_roots(
    apply: typing.Callable,
):
    """Two independent root nodes."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_id": [0, 1, 0, 1],
            }
        )
    )

    result = alifestd_try_add_ancestor_list_col_polars(df).lazy().collect()

    assert result["ancestor_list"].to_list() == [
        "[none]",
        "[none]",
        "[0]",
        "[1]",
    ]



import polars as pl
import pytest

from hstrat._auxiliary_lib import alifestd_check_topological_sensitivity_polars
from hstrat._auxiliary_lib._alifestd_check_topological_sensitivity import (
    _topologically_sensitive_cols,
    _update_only_sensitive_cols,
)


@pytest.fixture
def base_df():
    return pl.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 1],
            "origin_time": [0, 1, 2],
        }
    )


def test_none_present(base_df: pl.DataFrame):
    result = alifestd_check_topological_sensitivity_polars(
        base_df,
        insert=True,
        delete=True,
        update=True,
    )
    assert result == []


@pytest.mark.parametrize("col", sorted(_topologically_sensitive_cols))
def test_single_sensitive_col(base_df: pl.DataFrame, col: str):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=True,
        update=True,
    )
    assert result == [col]


def test_multiple_sensitive_cols(base_df: pl.DataFrame):
    df = base_df.with_columns(
        pl.lit(0).alias("branch_length"),
        pl.lit(0).alias("node_depth"),
        pl.lit(0).alias("sister_id"),
    )
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=True,
        update=True,
    )
    assert set(result) == {"branch_length", "node_depth", "sister_id"}


def test_non_sensitive_cols_ignored(base_df: pl.DataFrame):
    df = base_df.with_columns(
        pl.lit("x").alias("taxon_label"),
        pl.lit(True).alias("extant"),
    )
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=True,
        update=True,
    )
    assert result == []


def test_no_mutation(base_df: pl.DataFrame):
    df = base_df.with_columns(pl.lit(0).alias("branch_length"))
    original = df.clone()
    alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=True,
        update=True,
    )
    assert df.equals(original)


def test_empty_dataframe():
    df = pl.DataFrame({"id": pl.Series([], dtype=pl.Int64)})
    assert (
        alifestd_check_topological_sensitivity_polars(
            df,
            insert=True,
            delete=True,
            update=True,
        )
        == []
    )


def test_empty_dataframe_with_sensitive():
    df = pl.DataFrame(
        {
            "id": pl.Series([], dtype=pl.Int64),
            "branch_length": pl.Series([], dtype=pl.Float64),
        }
    )
    assert alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=True,
        update=True,
    ) == ["branch_length"]


@pytest.mark.parametrize("col", sorted(_topologically_sensitive_cols))
def test_lazyframe(base_df: pl.DataFrame, col: str):
    df = base_df.with_columns(pl.lit(0).alias(col))
    lazy = df.lazy()
    result = alifestd_check_topological_sensitivity_polars(
        lazy,
        insert=True,
        delete=True,
        update=True,
    )
    assert result == [col]


def test_lazyframe_none_present(base_df: pl.DataFrame):
    lazy = base_df.lazy()
    result = alifestd_check_topological_sensitivity_polars(
        lazy,
        insert=True,
        delete=True,
        update=True,
    )
    assert result == []


@pytest.mark.parametrize("col", sorted(_update_only_sensitive_cols))
def test_insert_only_excludes_update_only(base_df: pl.DataFrame, col: str):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=False,
        update=False,
    )
    assert col not in result


@pytest.mark.parametrize(
    "col",
    sorted(_topologically_sensitive_cols - _update_only_sensitive_cols),
)
def test_insert_only_includes_structure_sensitive(
    base_df: pl.DataFrame, col: str
):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=True,
        delete=False,
        update=False,
    )
    assert result == [col]


@pytest.mark.parametrize("col", sorted(_update_only_sensitive_cols))
def test_delete_only_excludes_update_only(base_df: pl.DataFrame, col: str):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=False,
        delete=True,
        update=False,
    )
    assert col not in result


@pytest.mark.parametrize(
    "col",
    sorted(_topologically_sensitive_cols - _update_only_sensitive_cols),
)
def test_delete_only_includes_structure_sensitive(
    base_df: pl.DataFrame, col: str
):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=False,
        delete=True,
        update=False,
    )
    assert result == [col]


@pytest.mark.parametrize("col", sorted(_topologically_sensitive_cols))
def test_update_includes_all(base_df: pl.DataFrame, col: str):
    df = base_df.with_columns(pl.lit(0).alias(col))
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=False,
        delete=False,
        update=True,
    )
    assert result == [col]


def test_no_ops_returns_empty(base_df: pl.DataFrame):
    cols = [pl.lit(0).alias(col) for col in _topologically_sensitive_cols]
    df = base_df.with_columns(cols)
    result = alifestd_check_topological_sensitivity_polars(
        df,
        insert=False,
        delete=False,
        update=False,
    )
    assert result == []


@pytest.mark.parametrize("col", sorted(_update_only_sensitive_cols))
def test_insert_only_lazyframe_excludes(base_df: pl.DataFrame, col: str):
    lazy = base_df.with_columns(pl.lit(0).alias(col)).lazy()
    result = alifestd_check_topological_sensitivity_polars(
        lazy,
        insert=True,
        delete=False,
        update=False,
    )
    assert col not in result



import typing


# adapted from https://github.com/mmore500/joinem/blob/v0.11.1/joinem/_dataframe_cli.py#L120
def eval_kwargs(kwargs_list: typing.List[str]) -> typing.Dict:
    """Parse a list of 'key=value' strings into a dictionary.

    Values are evaluated as Python expressions.

    Parameters
    ----------
    kwargs_list : list of str
        Each element should be a 'key=value' string.

    Returns
    -------
    dict
        Parsed keyword arguments.
    """
    to_eval = f"dict({','.join(kwargs_list)})"
    try:
        return eval(to_eval)
    except Exception as e:
        raise ValueError(
            f"Failed to parse kwarg expressions {kwargs_list!r} "
            f"via {to_eval!r}",
        ) from e



import polars as pl

from hstrat._auxiliary_lib import alifestd_count_inner_nodes_polars


def test_empty_df():
    df = pl.DataFrame(
        {"id": [], "ancestor_id": []},
        schema={"id": pl.Int64, "ancestor_id": pl.Int64},
    )
    assert alifestd_count_inner_nodes_polars(df) == 0


def test_singleton_df():
    df = pl.DataFrame(
        {
            "id": [0],
            "ancestor_id": [0],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 0


def test_one_root_two_children():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 0],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 1


def test_strictly_bifurcating_df():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_id": [0, 0, 0, 1, 1],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 2


def test_polytomy_df():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_id": [0, 0, 0, 0, 1],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 2


def test_multiple_trees_df1():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [0, 1, 0, 2, 2, 3],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 3


def test_multiple_trees_df2():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [0, 1, 0, 1, 2, 3],
        }
    )
    assert alifestd_count_inner_nodes_polars(df) == 4



import argparse
import contextlib
import functools
import gc
import logging
import os
import sys
import typing

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import opytional as opyt
import polars as pl

from ._RngStateContext import RngStateContext
from ._add_bool_arg import add_bool_arg
from ._alifestd_find_leaf_ids_polars import alifestd_find_leaf_ids_polars
from ._alifestd_prune_extinct_lineages_polars import (
    alifestd_prune_extinct_lineages_polars,
)
from ._alifestd_topological_sensitivity_warned_polars import (
    alifestd_topological_sensitivity_warned_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration
from ._log_memory_usage import log_memory_usage


def _alifestd_downsample_tips_polars_impl(
    phylogeny_df: pl.DataFrame,
    n_downsample: int,
) -> pl.DataFrame:
    """Implementation detail for alifestd_downsample_tips_polars."""

    logging.info(
        "- alifestd_downsample_tips_polars: collecting leaf ids...",
    )
    leaf_ids = alifestd_find_leaf_ids_polars(phylogeny_df)
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_polars: sampling leaf_ids...",
    )
    leaf_ids = np.random.choice(
        leaf_ids, size=min(n_downsample, len(leaf_ids)), replace=False
    )
    gc.collect()
    log_memory_usage(logging.info)

    logging.info("- alifestd_downsample_tips_polars: collecting len(df)...")
    len_df = phylogeny_df.lazy().select(pl.len()).collect().item()

    logging.info("- alifestd_downsample_tips_polars: finding extant...")
    extant_mask = np.bincount(leaf_ids, minlength=len_df).astype(bool)
    del leaf_ids
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_polars: marking extant...",
    )
    phylogeny_df = phylogeny_df.with_columns(extant=extant_mask)
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_polars: pruning...",
    )
    return alifestd_prune_extinct_lineages_polars(phylogeny_df).drop("extant")


@alifestd_topological_sensitivity_warned_polars(
    insert=False,
    delete=True,
    update=False,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_downsample_tips_polars instead.",
)
def alifestd_downsample_tips_polars(
    phylogeny_df: pl.DataFrame,
    n_downsample: int,
    seed: typing.Optional[int] = None,
) -> pl.DataFrame:
    """Create a subsample phylogeny containing `n_downsample` tips.

    If `n_downsample` is greater than the number of tips in the phylogeny,
    the whole phylogeny is returned.

    Only supports asexual phylogenies.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    n_downsample : int
        Number of tips to retain.
    seed : int, optional
        Integer seed for deterministic behavior.

    Raises
    ------
    NotImplementedError
        If `phylogeny_df` has no "ancestor_id" column.

    Returns
    -------
    polars.DataFrame
        The downsampled phylogeny in alife standard format.

    See Also
    --------
    alifestd_downsample_tips_asexual :
        Pandas-based implementation.
    """
    if "ancestor_id" not in phylogeny_df.lazy().collect_schema().names():
        raise NotImplementedError("ancestor_id column required")

    if phylogeny_df.lazy().limit(1).collect().is_empty():
        return phylogeny_df

    with opyt.apply_if_or_else(seed, RngStateContext, contextlib.nullcontext):
        return _alifestd_downsample_tips_polars_impl(
            phylogeny_df,
            n_downsample,
        )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Create a subsample phylogeny containing `-n` tips.

If `-n` is greater than the number of tips in the phylogeny, the whole phylogeny is returned.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_downsample_tips_asexual :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_downsample_tips_polars",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "-n",
        default=sys.maxsize,
        type=int,
        help="Number of tips to subsample.",
    )
    parser.add_argument(
        "--seed",
        default=None,
        dest="seed",
        help="Integer seed for deterministic behavior.",
        type=int,
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_downsample_tips_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=functools.partial(
                    alifestd_downsample_tips_polars,
                    n_downsample=args.n,
                    seed=args.seed,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
                overridden_arguments="ignore",  # seed is overridden
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



import os

from downstream import dstream, dsurf
from phyloframe import legacy as pfl
import polars as pl

from hstrat import hstrat
from hstrat.dataframe import (
    surface_postprocess_trie,
    surface_unpack_reconstruct,
)
from hstrat.phylogenetic_inference.tree.trie_postprocess import (
    AssignOriginTimeNodeRankTriePostprocessor,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_smoke():
    df = pl.read_csv(f"{assets_path}/packed.csv")
    raw = surface_unpack_reconstruct(df)
    res = surface_postprocess_trie(
        raw,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(
            t0="dstream_S",
        ),
    )
    assert "origin_time" in res.columns
    assert len(res) <= len(raw)
    assert pfl.alifestd_validate(
        pfl.alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert pfl.alifestd_is_chronologically_ordered(res.to_pandas())


def test_dstream_rank_in_unpack_reconstruct():
    """dstream_rank should be present after surface_unpack_reconstruct."""
    df = pl.read_csv(f"{assets_path}/packed.csv")
    raw = surface_unpack_reconstruct(df)
    assert "dstream_rank" in raw.columns


def test_zero_generations_elapsed():
    """Postprocessing should handle surfaces with zero generations elapsed.

    When dstream_T == dstream_S, no generations have elapsed beyond the
    founding ancestor's initial surface fill. This should not cause errors
    or produce an empty result.
    """
    S = 8
    differentia_bitwidth = 8
    dstream_T_bitwidth = 8
    algo = dstream.steady_algo

    # Create surfaces where dstream_T == dstream_S (zero generations elapsed)
    surfaces = [
        hstrat.HereditaryStratigraphicSurface(
            dsurf.Surface(algo, S),
            predeposit_strata=0,
            stratum_differentia_bit_width=differentia_bitwidth,
        )
        for _ in range(3)
    ]
    # Deposit exactly S strata so dstream_T == S (zero generations elapsed)
    for surf in surfaces:
        surf.DepositStrata(S)

    df = pl.DataFrame(
        [
            {
                "data_hex": hstrat.surf_to_hex(
                    surf, dstream_T_bitwidth=dstream_T_bitwidth
                ),
                "dstream_algo": f"dstream.{algo.__name__.split('.')[-1]}",
                "dstream_storage_bitoffset": dstream_T_bitwidth,
                "dstream_storage_bitwidth": surf.S * differentia_bitwidth,
                "dstream_T_bitoffset": 0,
                "dstream_T_bitwidth": dstream_T_bitwidth,
                "dstream_S": surf.S,
            }
            for surf in surfaces
        ]
    )
    raw = surface_unpack_reconstruct(df)
    res = surface_postprocess_trie(
        raw,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(
            t0="dstream_S",
        ),
    )
    assert len(res) > 0
    assert "origin_time" in res.columns
    assert pfl.alifestd_validate(
        pfl.alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_try_add_ancestor_id_col_polars instead.",
)
def alifestd_try_add_ancestor_id_col_polars(
    phylogeny_df: pl.DataFrame,
) -> pl.DataFrame:
    """Add an ancestor_id column to the input DataFrame if the phylogeny is
    asexual and the column does not already exist.
    """

    if "ancestor_id" in phylogeny_df.lazy().collect_schema().names() or (
        phylogeny_df.lazy()
        .select(
            pl.col("ancestor_list").str.contains(",").any(),
        )
        .collect()
        .item()
    ):
        return phylogeny_df

    return phylogeny_df.with_columns(
        ancestor_id=pl.col("ancestor_list")
        .str.extract(r"(\d+)", 1)
        .cast(pl.Int64, strict=False)
        .fill_null(pl.col("id").cast(pl.Int64))
    )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Add an ancestor_id column if the phylogeny is asexual and the column does not already exist.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_try_add_ancestor_id_col_polars",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_try_add_ancestor_id_col_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=alifestd_try_add_ancestor_id_col_polars,
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



import argparse
import functools
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned(
    insert=True,
    delete=False,
    update=True,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_add_inner_knuckles_asexual instead.",
)
def alifestd_add_inner_knuckles_asexual(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """For all inner nodes, add a subtending unifurcation ("knuckle").

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_leaves(phylogeny_df, mutate=True)

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    if "ancestor_id" not in phylogeny_df:
        raise ValueError("asexual phylogeny required")

    inner_df = phylogeny_df[~phylogeny_df["is_leaf"]].copy()

    if inner_df.empty:
        return phylogeny_df

    if "is_root" in inner_df:
        inner_df["is_root"] = False

    if "origin_time_delta" in inner_df:
        inner_df["origin_time_delta"] = 0

    id_delta = phylogeny_df["id"].max() + 1

    inner_df["ancestor_id"] += id_delta * (
        inner_df["id"] == inner_df["ancestor_id"]
    )
    inner_df["id"] += id_delta

    if not (inner_df["id"].min() > phylogeny_df["id"].max()):
        raise ValueError("overflow in new id assigment")

    phylogeny_df.loc[~phylogeny_df["is_leaf"], "ancestor_id"] = (
        phylogeny_df.loc[~phylogeny_df["is_leaf"], "id"] + id_delta
    )

    res = pd.concat([phylogeny_df, inner_df], ignore_index=True)

    if "ancestor_list" in res:
        res["ancestor_list"] = alifestd_make_ancestor_list_col(
            res["id"],
            res["ancestor_id"],
        )

    return res


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

For all inner nodes, add a subtending unifurcation (knuckle).

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_add_inner_knuckles_asexual",
        dfcli_version=get_hstrat_version(),
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_add_inner_knuckles_asexual",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_add_inner_knuckles_asexual,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
        )



import argparse
import functools
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
from tqdm import tqdm

from hstrat.dataframe._surface_test_drive import surface_test_drive

from .._auxiliary_lib import (
    begin_prod_logging,
    format_cli_description,
    get_hstrat_version,
    log_context_duration,
)

raw_message = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Reads alife standard phylogeny dataframe to create a population of hstrat surface annotations corresponding to the phylogeny tips, "as-if" they had evolved according to the provided phylogeny history.

Parameters
----------
df : pl.DataFrame
    The input DataFrame containing alife standard phylogeny with required columns, one row per taxon.

    Note that the alife-standard `ancestor_list` column is not required.

    Required schema:
        - 'id' : pl.UInt64
            - Taxon identifier.
        - 'ancestor_id' : pl.UInt64
            - Taxon identifier of ancestor.
            - Own 'id' if root.

    Optional schema:
        - 'origin_time' : pl.UInt64
            - Number of generations elapsed from ancestor.
            - Determines branch lengths.
            - Otherwise, all branches are assumed to be length 1.
        - 'extant' : pl.Boolean
            - Should an entry corresponding to this phylogeny taxon be included in the output population?
            - Otherwise, all tips are considered extant and all inner nodes are not.
        - Additional user-defined columns will be forwarded to the output DataFrame.

dstream_algo : str
    Name of downstream curation algorithm to use.

dstream_S : int
    Capacity of annotation dstream buffer, in number of data items.

stratum_differentia_bit_width : int
    The bit width of the generated differentia.

Returns
-------
pl.DataFrame
    The output DataFrame containing generated hstrat surface annotations.

    Required schema:
        - 'data_hex' : pl.String
            - Raw genome data, with serialized dstream buffer and counter.
            - Represented as a hexadecimal string.
        - 'dstream_algo' : pl.Categorical
            - Name of downstream curation algorithm used.
            - e.g., 'dstream.steady_algo'
        - 'dstream_storage_bitoffset' : pl.UInt64
            - Position of dstream buffer field in 'data_hex'.
        - 'dstream_storage_bitwidth' : pl.UInt64
            - Size of dstream buffer field in 'data_hex'.
        - 'dstream_T_bitoffset' : pl.UInt64
            - Position of dstream counter field in 'data_hex'.
        - 'dstream_T_bitwidth' : pl.UInt64
            - Size of dstream counter field in 'data_hex'.
        - 'dstream_S' : pl.Uint32
            - Capacity of dstream buffer, in number of data items.
        - 'origin_time' : pl.UInt64
            - Number of generations elapsed since the founding ancestor.

    Additional user-defined columns will be forwarded from the input DataFrame.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(raw_message),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_parser_base(
        parser=parser,
        dfcli_module="hstrat.dataframe.surface_test_drive",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--stratum-differentia-bit-width",
        type=int,
        default=1,
        help="Number of bits used per differentia.",
    )
    parser.add_argument(
        "--dstream-algo",
        type=str,
        default="dstream.tiltedxtc_algo",
        help="Downstream algorithm to curate differentia retention.",
    )
    parser.add_argument(
        "--dstream-S",
        type=int,
        default=64,
        help="Number differentiae to store per surface (usually power of 2).",
    )
    parser.add_argument(
        "--dstream-T-bitwidth",
        type=int,
        default=32,
        help="Number of bits for generation (dstream T) counter.",
    )
    return parser


def _main() -> None:
    parser = _create_parser()
    args, __ = parser.parse_known_args()

    with log_context_duration(
        "hstrat.dataframe.surface_test_drive", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=functools.partial(
                surface_test_drive,
                stratum_differentia_bit_width=args.stratum_differentia_bit_width,
                dstream_algo=args.dstream_algo,
                dstream_S=args.dstream_S,
                progress_wrap=tqdm,
            ),
        )


if __name__ == "__main__":
    begin_prod_logging()
    _main()



import itertools as it
import typing

from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd

from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_has_multiple_roots import alifestd_has_multiple_roots
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_is_working_format_asexual import (
    alifestd_is_working_format_asexual,
)
from ._alifestd_mark_num_descendants_asexual import (
    alifestd_mark_num_descendants_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit


@jit(nopython=True)
def _alifestd_unfurl_traversal_semiorder_asexual_fast_path(
    ancestor_ids: np.ndarray,
    num_descendants: np.ndarray,
    leaf_ids: np.ndarray,
) -> np.ndarray:
    """Implementation detail for phylogenies in working format."""
    dtype = ancestor_ids.dtype
    ancestor_lookup = ancestor_ids
    num_descendants_lookup = num_descendants

    site_assignments = np.zeros(len(ancestor_ids), dtype=dtype)
    visited = np.zeros(len(ancestor_ids), dtype=np.bool_)
    for leaf_id in leaf_ids:

        # find offset of closest visited ancestor, if any
        current_id = leaf_id
        while True:
            if visited[current_id]:
                offset = site_assignments[current_id] + 1
                break

            current_id, prev_id = ancestor_lookup[current_id], current_id
            if current_id == prev_id:
                offset = 0
                break

        site_assignments[leaf_id] = offset
        prev_id, current_id = leaf_id, ancestor_lookup[leaf_id]
        while current_id != prev_id:
            if visited[current_id]:
                break
            visited[current_id] = True

            assigned_site = offset + num_descendants_lookup[prev_id] + 1
            site_assignments[current_id] = assigned_site

            current_id, prev_id = ancestor_lookup[current_id], current_id

    result = np.empty(len(ancestor_ids), dtype=dtype)
    result[site_assignments] = np.arange(site_assignments.size)
    return result


def _alifestd_unfurl_traversal_semiorder_asexual_slow_path(
    phylogeny_df: pd.DataFrame,
    leaf_ids: np.ndarray,
) -> np.ndarray:
    """Implementation detail for phylogenies not in working format."""
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    phylogeny_df = alifestd_mark_num_descendants_asexual(
        phylogeny_df, mutate=True
    )

    ancestor_lookup = dict(
        zip(phylogeny_df["id"], phylogeny_df["ancestor_id"]),
    )
    num_descendants_lookup = dict(
        zip(phylogeny_df["id"], phylogeny_df["num_descendants"]),
    )

    def iter_lineage(leaf_id: int) -> typing.Iterator[int]:
        current_id = leaf_id
        while True:
            yield current_id
            current_id, prev_id = ancestor_lookup[current_id], current_id
            if current_id == prev_id:
                break

    site_assignments = dict()
    visited = set()
    for leaf_id in leaf_ids:

        # find offset of closest visited ancestor, if any
        visited_ancestor = next(
            filter(visited.__contains__, iter_lineage(leaf_id)),
            None,
        )
        offset = (
            0
            if visited_ancestor is None
            else site_assignments[visited_ancestor] + 1
        )

        site_assignments[leaf_id] = offset
        for prev_id, current_id in it.pairwise(
            it.takewhile(
                lambda lineage_id: lineage_id not in visited,
                iter_lineage(leaf_id),
            ),
        ):
            visited.add(current_id)

            assigned_site = offset + num_descendants_lookup[prev_id] + 1
            assert 0 <= assigned_site < len(phylogeny_df)
            assert assigned_site not in site_assignments.values()

            assert current_id not in site_assignments
            site_assignments[current_id] = assigned_site

        visited.add(leaf_id)

    assert len(site_assignments) == len(phylogeny_df)
    assert len(site_assignments) == len(visited)
    assert len(site_assignments) == len(set(site_assignments.values()))

    result = np.empty(len(phylogeny_df), dtype=int)
    for id_, assigned_site in site_assignments.items():
        result[assigned_site] = id_
    return result


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_unfurl_traversal_semiorder_asexual instead.",
)
def alifestd_unfurl_traversal_semiorder_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> np.ndarray:
    """List `id` values in semiorder traversal order.

    An inorder traversal where either left child (smaller id) or right child
    (larger id) may be visited first.

    The provided dataframe must be asexual and strictly bifurcating.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if alifestd_has_multiple_roots(phylogeny_df):
        raise ValueError(
            "Phylogeny must have a single root for inorder traversal."
        )
    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError(
            "Phylogeny must be strictly bifurcating for inorder traversal.",
        )

    if alifestd_is_working_format_asexual(phylogeny_df, mutate=True):
        if "num_descendants" not in phylogeny_df.columns:
            phylogeny_df = alifestd_mark_num_descendants_asexual(
                phylogeny_df, mutate=True
            )
        return _alifestd_unfurl_traversal_semiorder_asexual_fast_path(
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["num_descendants"].to_numpy(),
            alifestd_find_leaf_ids(phylogeny_df),
        )
    else:
        return _alifestd_unfurl_traversal_semiorder_asexual_slow_path(
            phylogeny_df,
            alifestd_find_leaf_ids(phylogeny_df),
        )



import os
import typing

import pandas as pd
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_mark_num_children_asexual,
    alifestd_to_working_format,
)
from hstrat._auxiliary_lib._alifestd_mark_num_children_polars import (
    alifestd_mark_num_children_polars,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(
                f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
            )
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_fuzz(
    phylogeny_df: pd.DataFrame, apply: typing.Callable
):
    """Verify num_children column is correctly added."""
    df_prepared = pl.from_pandas(phylogeny_df)
    df_pl = apply(df_prepared)

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert "num_children" in result.columns
    assert len(result) == len(df_prepared)

    # every id in result should match original
    assert result["id"].to_list() == df_prepared["id"].to_list()

    # all counts should be non-negative
    assert (result["num_children"] >= 0).all()

    # all counts should be less than total number of nodes
    assert (result["num_children"] < len(result)).all()

    # leaves (not appearing as ancestor of any non-root node) should have 0
    internal_ids = set(
        df_prepared.filter(pl.col("ancestor_id") != pl.col("id"))
        .select("ancestor_id")
        .to_series()
        .to_list()
    )
    for row in result.iter_rows(named=True):
        if row["id"] not in internal_ids:
            assert row["num_children"] == 0
        else:
            assert row["num_children"] >= 1


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(
                f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
            )
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_matches_pandas(
    phylogeny_df: pd.DataFrame, apply: typing.Callable
):
    """Verify polars result matches pandas result."""
    result_pd = alifestd_mark_num_children_asexual(phylogeny_df, mutate=False)

    df_pl = apply(pl.from_pandas(phylogeny_df))
    result_pl = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result_pd["num_children"].tolist() == (
        result_pl["num_children"].to_list()
    )


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_simple_chain(
    apply: typing.Callable,
):
    """Test a simple chain: 0 -> 1 -> 2."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [1, 1, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_simple_tree(apply: typing.Callable):
    """Test a simple tree.

    Tree structure:
        0 (root)
        +-- 1
        |   +-- 3
        |   +-- 4
        +-- 2
    """
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3, 4],
                "ancestor_id": [0, 0, 0, 1, 1],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [2, 2, 0, 0, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_two_roots(apply: typing.Callable):
    """Two independent roots with children."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_id": [0, 1, 0, 1],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [1, 1, 0, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_all_roots(apply: typing.Callable):
    """All self-referencing roots have 0 children."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 1, 2],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [0, 0, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_single_node(apply: typing.Callable):
    """A single root node has 0 children."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0],
                "ancestor_id": [0],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_empty(apply: typing.Callable):
    """Empty dataframe gets num_children column."""
    df_pl = apply(
        pl.DataFrame(
            {"id": [], "ancestor_id": []},
            schema={"id": pl.Int64, "ancestor_id": pl.Int64},
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert "num_children" in result.columns
    assert result.is_empty()


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_preserves_columns(
    apply: typing.Callable,
):
    """Verify original columns are preserved."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
                "origin_time": [0.0, 1.0, 2.0],
                "taxon_label": ["a", "b", "c"],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert "num_children" in result.columns
    assert "origin_time" in result.columns
    assert "taxon_label" in result.columns
    assert result["origin_time"].to_list() == [0.0, 1.0, 2.0]
    assert result["taxon_label"].to_list() == ["a", "b", "c"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_with_preexisting_is_root(
    apply: typing.Callable,
):
    """If is_root already exists, it should be used directly."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
                "is_root": [True, False, False],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [1, 1, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_non_contiguous_ids(
    apply: typing.Callable,
):
    """Verify NotImplementedError for non-contiguous ids."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 2, 5],
                "ancestor_id": [0, 0, 2],
            }
        ),
    )
    with pytest.raises(NotImplementedError):
        alifestd_mark_num_children_polars(df_pl).lazy().collect()


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_unsorted(apply: typing.Callable):
    """Verify NotImplementedError for topologically unsorted data."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 2, 0],
            }
        ),
    )
    with pytest.raises(NotImplementedError):
        alifestd_mark_num_children_polars(df_pl).lazy().collect()


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_mark_num_children_polars_star_topology(
    apply: typing.Callable,
):
    """Root with many direct children."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5],
                "ancestor_id": [0, 0, 0, 0, 0, 0],
            }
        ),
    )

    result = alifestd_mark_num_children_polars(df_pl).lazy().collect()

    assert result["num_children"].to_list() == [5, 0, 0, 0, 0, 0]



import os

from phyloframe import legacy as pfl
import polars as pl

from hstrat.dataframe import surface_build_tree
from hstrat.phylogenetic_inference.tree.trie_postprocess import (
    AssignOriginTimeNodeRankTriePostprocessor,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_smoke():
    df = pl.read_csv(f"{assets_path}/packed.csv")
    res = surface_build_tree(
        df,
        collapse_unif_freq=0,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(
            t0="dstream_S",
        ),
    )
    assert "origin_time" in res.columns
    assert len(df) <= len(res)
    assert pfl.alifestd_validate(
        pfl.alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert pfl.alifestd_is_chronologically_ordered(res.to_pandas())


def _get_full_schema():
    """Get expected output schema from a known-good two-genome run."""
    df = pl.read_csv(f"{assets_path}/packed.csv")
    return surface_build_tree(df, collapse_unif_freq=0).schema


def test_empty():
    """Regression: empty genome set should produce an empty phylogeny without
    crashing, with the same columns and types as a non-empty result."""
    df = pl.read_csv(f"{assets_path}/packed.csv").head(0)
    res = surface_build_tree(df)
    assert len(res) == 0
    full_schema = _get_full_schema()
    assert set(res.schema.names()) == set(full_schema.names())
    for col in full_schema.names():
        assert (
            res.schema[col] == full_schema[col]
        ), f"type mismatch for {col}: {res.schema[col]} != {full_schema[col]}"


def test_single_genome():
    """Regression: a single genome should not crash."""
    df = pl.read_csv(f"{assets_path}/packed.csv").head(1)
    res = surface_build_tree(
        df,
        collapse_unif_freq=0,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(
            t0="dstream_S",
        ),
    )
    # single genome may produce empty tree after trunk deletion
    assert isinstance(res, pl.DataFrame)
    if len(res) == 0:
        full_schema = _get_full_schema()
        assert set(res.schema.names()) == set(full_schema.names())
        for col in full_schema.names():
            assert res.schema[col] == full_schema[col], (
                f"type mismatch for {col}: "
                f"{res.schema[col]} != {full_schema[col]}"
            )


def test_single_genome_no_delete_trunk():
    """Regression: a single genome without trunk deletion should produce a
    valid phylogeny."""
    df = pl.read_csv(f"{assets_path}/packed.csv").head(1)
    res = surface_build_tree(
        df,
        collapse_unif_freq=0,
        delete_trunk=False,
        trie_postprocessor=AssignOriginTimeNodeRankTriePostprocessor(
            t0="dstream_S",
        ),
    )
    assert len(res) >= 1
    assert pfl.alifestd_validate(
        pfl.alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert pfl.alifestd_is_chronologically_ordered(res.to_pandas())



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_check_topological_sensitivity import (
    alifestd_check_topological_sensitivity,
)
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_drop_topological_sensitivity instead.",
)
def alifestd_drop_topological_sensitivity(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    insert: bool = True,
    delete: bool = True,
    update: bool = True,
) -> pd.DataFrame:
    """Drop columns from `phylogeny_df` that may be invalidated by
    topological operations such as collapsing unifurcations.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.
    mutate : bool, default False
        Are side effects on the input argument allowed?
    insert : bool, default True
        Drop columns sensitive to node insertion.
    delete : bool, default True
        Drop columns sensitive to node deletion.
    update : bool, default True
        Drop columns sensitive to ancestor relationship updates.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_drop_topological_sensitivity_polars :
        Polars-based implementation.
    """
    to_drop = alifestd_check_topological_sensitivity(
        phylogeny_df,
        insert=insert,
        delete=delete,
        update=update,
    )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df.drop(columns=to_drop, inplace=True)
    return phylogeny_df


_raw_description = f"""\
{os.path.basename(__file__)} | \
(hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Drop columns that may be invalidated by topological operations.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_drop_topological_sensitivity_polars :
    Entrypoint for high-performance Polars-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module=(
            "hstrat._auxiliary_lib" "._alifestd_drop_topological_sensitivity"
        ),
        dfcli_version=get_hstrat_version(),
    )
    add_bool_arg(
        parser,
        "insert",
        default=True,
        help="drop columns sensitive to node insertion (default: True)",
    )
    add_bool_arg(
        parser,
        "delete",
        default=True,
        help="drop columns sensitive to node deletion (default: True)",
    )
    add_bool_arg(
        parser,
        "update",
        default=True,
        help=(
            "drop columns sensitive to ancestor relationship updates"
            " (default: True)"
        ),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib" "._alifestd_drop_topological_sensitivity",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                lambda df: alifestd_drop_topological_sensitivity(
                    df,
                    insert=args.insert,
                    delete=args.delete,
                    update=args.update,
                ),
            ),
        )



import logging
import os
import subprocess
import typing


# adapted from https://stackoverflow.com/a/47411778/17332200
def log_memory_usage(logger: typing.Callable = logging.info) -> None:
    """Log memory use."""
    if "HSTRAT_LOG_MEMORY_USAGE" in os.environ:
        try:
            message = (
                "memory usage:\n"
                + subprocess.check_output(
                    [
                        "free",
                        "--human",
                        "--total",
                    ],
                    shell=False,
                ).decode()
            )
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            message = f"logging memory use failed: {e}"

        logger(message)



from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd

from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_has_multiple_roots import alifestd_has_multiple_roots
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_is_working_format_asexual import (
    alifestd_is_working_format_asexual,
)
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_mark_num_descendants_asexual import (
    alifestd_mark_num_descendants_asexual,
)
from ._alifestd_mark_num_preceding_leaves_asexual import (
    alifestd_mark_num_preceding_leaves_asexual,
)
from ._alifestd_unfurl_traversal_semiorder_asexual import (
    _alifestd_unfurl_traversal_semiorder_asexual_fast_path,
    _alifestd_unfurl_traversal_semiorder_asexual_slow_path,
)


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_unfurl_traversal_inorder_asexual instead.",
)
def alifestd_unfurl_traversal_inorder_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> np.ndarray:
    """List `id` values in semiorder traversal order, with left children
    visited first.

    The provided dataframe must be asexual and strictly bifurcating.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if alifestd_has_multiple_roots(phylogeny_df):
        raise ValueError(
            "Phylogeny must have a single root for inorder traversal."
        )
    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError(
            "Phylogeny must be strictly bifurcating for inorder traversal.",
        )

    if "num_preceding_leaves" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_preceding_leaves_asexual(
            phylogeny_df,
            mutate=True,
        )

    if alifestd_is_working_format_asexual(phylogeny_df, mutate=True):
        if "num_descendants" not in phylogeny_df.columns:
            phylogeny_df = alifestd_mark_num_descendants_asexual(
                phylogeny_df, mutate=True
            )
        leaf_positions = phylogeny_df["num_preceding_leaves"].to_numpy()

        leaf_ids = alifestd_find_leaf_ids(phylogeny_df)
        sorted_leaf_ids = np.empty_like(leaf_ids)
        sorted_leaf_ids[leaf_positions[leaf_ids]] = leaf_ids

        return _alifestd_unfurl_traversal_semiorder_asexual_fast_path(
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["num_descendants"].to_numpy(),
            sorted_leaf_ids,
        )
    else:
        if "is_leaf" not in phylogeny_df.columns:
            phylogeny_df = alifestd_mark_leaves(phylogeny_df, mutate=True)

        sorted_phylogeny_df = phylogeny_df.sort_values("num_preceding_leaves")
        sorted_leaf_ids = sorted_phylogeny_df.loc[
            sorted_phylogeny_df["is_leaf"], "id"
        ]

        return _alifestd_unfurl_traversal_semiorder_asexual_slow_path(
            phylogeny_df,
            sorted_leaf_ids,
        )



import os
import typing

import pandas as pd
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_mark_node_depth_asexual,
    alifestd_to_working_format,
)
from hstrat._auxiliary_lib._alifestd_mark_node_depth_polars import (
    alifestd_mark_node_depth_polars,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(
                f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
            )
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_fuzz_matches_pandas(
    phylogeny_df: pd.DataFrame, apply: typing.Callable
):
    """Verify polars node_depth matches pandas implementation on real data."""
    result_pd = alifestd_mark_node_depth_asexual(phylogeny_df, mutate=False)

    df_pl = apply(pl.from_pandas(phylogeny_df))
    result_pl = alifestd_mark_node_depth_polars(df_pl).lazy().collect()

    expected = result_pd["node_depth"].tolist()
    actual = result_pl["node_depth"].to_list()
    assert actual == expected


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_simple_chain(apply: typing.Callable):
    """Test a simple chain: 0 -> 1 -> 2."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert result["node_depth"].to_list() == [0, 1, 2]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_simple_tree(apply: typing.Callable):
    """Test: 0 (root) -> 1, 0 -> 2, 1 -> 3."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_id": [0, 0, 0, 1],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert result["node_depth"].to_list() == [0, 1, 1, 2]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_multiple_roots(apply: typing.Callable):
    """Each independent root should have depth 0."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 1, 2],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert result["node_depth"].to_list() == [0, 0, 0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_single_node(apply: typing.Callable):
    """A single root node should have depth 0."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0],
                "ancestor_id": [0],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert result["node_depth"].to_list() == [0]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_empty(apply: typing.Callable):
    """Empty dataframe gets node_depth column."""
    df_pl = apply(
        pl.DataFrame(
            {"id": [], "ancestor_id": []},
            schema={"id": pl.Int64, "ancestor_id": pl.Int64},
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert "node_depth" in result.columns
    assert result.is_empty()


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_preserves_columns(apply: typing.Callable):
    """Verify original columns are preserved and node_depth is added."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
                "origin_time": [0.0, 1.0, 2.0],
                "taxon_label": ["a", "b", "c"],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert "node_depth" in result.columns
    assert "origin_time" in result.columns
    assert "taxon_label" in result.columns
    assert result["origin_time"].to_list() == [0.0, 1.0, 2.0]
    assert result["taxon_label"].to_list() == ["a", "b", "c"]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_with_ancestor_list_col(apply: typing.Callable):
    """Test that ancestor_list is correctly converted to ancestor_id."""
    df_pl = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_list": ["[None]", "[0]", "[0]", "[1]"],
            }
        )
    )
    result = alifestd_mark_node_depth_polars(df_pl).lazy().collect()
    assert result["node_depth"].to_list() == [0, 1, 1, 2]


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_does_not_mutate_input(apply: typing.Callable):
    """Verify the input dataframe is not mutated."""
    df_pl = pl.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 0],
        }
    )
    original_data = df_pl.clone()

    alifestd_mark_node_depth_polars(apply(df_pl))

    assert df_pl.columns == original_data.columns
    assert df_pl.equals(original_data)
